package cz.muni.fi.pb162.hw01.impl;

import cz.muni.fi.pb162.hw01.DefaultPortResolver;

import cz.muni.fi.pb162.hw01.url.SmartUrl;

import java.util.Arrays;
import java.util.ArrayList;

/**
 * @author Richard Ondrejka
 */
public class Url implements SmartUrl {
    private String rawUrl, canonicalUrl, protocol, host, path, query, fragment;
    private Integer port;
    private DefaultPortResolver portResolver = new DefaultPortResolver();
    private Boolean isCorrectUrl = true;

    /**
     * Constructor.
     * @param url url
     */
    public Url(String url) {
        rawUrl = url;
        process();
    }

    /**
     * Set parameters.
     */
    private void process() {
        String[] urlSplit = rawUrl.split("/");
        findProtocol(urlSplit);
        findPort(urlSplit);
        findHost(urlSplit);
        findPath(urlSplit);
        findFragment(urlSplit);
        findQuery(urlSplit);
        createCanonicalUrl();
    }

    /**
     * Create and sets canonical url.
     */
    private void createCanonicalUrl() {
        if (!isCorrectUrl) {
            canonicalUrl = null;
            return;
        }

        canonicalUrl = protocol + "://" + host;

        if (port != portResolver.getPort(protocol)) {
            canonicalUrl += ":" + port;
        }

        if (!path.equals("")){
            canonicalUrl += "/" + path;
        }

        if (!query.equals("")) {
            canonicalUrl += "?" + query;
        }

        if (!fragment.equals("")) {
            canonicalUrl += "#" + fragment;
        }
    }

    /**
     * Sort query parameters.
     */
    private String querySort(String query) {
        String[] sortedQuery = query.split("&");
        Arrays.sort(sortedQuery);
        for (int i = 0; i < sortedQuery.length; ++i) {
            if (i == 0) {
                query = sortedQuery[i];
            } else {
                query += "&" + sortedQuery[i];
            }
        }
        return query;
    }

    /**
     * Find query parameters.
     * @param urlSplit url
     */
    private void findQuery(String[] urlSplit) {
        String[] fragmentSplit = urlSplit[urlSplit.length - 1].split("\\?");
        if (fragmentSplit.length < 2) {
            query = "";
            return;
        }

        if (fragmentSplit[1].split("#").length > 0) {
            query = fragmentSplit[1].split("#")[0];
        } else {
            query = fragmentSplit[1];
        }
        query = querySort(query);
    }

    /**
     * Finds fragment parameter.
     * @param urlSplit url
     */
    private void findFragment(String[] urlSplit) {
        if (urlSplit[urlSplit.length - 1].split("#").length > 1) {
            fragment = urlSplit[urlSplit.length - 1].split("#")[1];
            return;
        }
        fragment = "";
    }

    /**
     * Removes query from string.
     * @param string string
     * @return string without query
     */
    private String removeQueryFromElement(String string){
        if (string.split("\\?").length > 0){
            return string.split("\\?")[0];
        }
        return string;
    }

    /**
     * Remove fragment from string.
     * @param string string
     * @return string without fragment
     */
    private String removeFragmentFromElement(String string){
        if (string.split("#").length > 0){
            return string.split("#")[0];
        }
        return string;
    }

    /**
     * Set path.
     * @param arrayList list
     */
    private void setPathFromList(ArrayList<String> arrayList){
        path = "";
        for (String string : arrayList){
            path += string;
        }
    }

    /**
     *Find path in url.
     * @param urlSplit url
     */
    private void findPath(String[] urlSplit) {
        ArrayList<String> pathList = new ArrayList<>();
        int listIndex = -1;
        for (int i = 3; i < urlSplit.length; ++i){
            switch (urlSplit[i]){
                case ".":
                    break;
                case "":
                    break;
                case "..":
                    if (pathList.size() == 0){
                        isCorrectUrl = false;
                        path = "";
                        return;
                    }
                    pathList.remove(pathList.size() - 1);
                    listIndex -= 1;
                    break;
                default:
                    if (i == urlSplit.length - 1){
                        pathList.add(removeFragmentFromElement(removeQueryFromElement(urlSplit[i])));
                    } else {
                        pathList.add(urlSplit[i]);
                    }

                    listIndex += 1;

                    if (pathList.size() > 1 && !pathList.get(listIndex).equals("")){
                        pathList.set(listIndex, "/" + pathList.get(listIndex));
                    }
            }
        }
        setPathFromList(pathList);
    }

    /**
     * Find protocol in url.
     * @param urlSplit url
     */
    private void findProtocol(String[] urlSplit) {
        if ("".equals(urlSplit[0]) || urlSplit[0].split(":").length != 1) {
            isCorrectUrl = false;
            protocol = null;
            return;
        }
        protocol = urlSplit[0].substring(0, urlSplit[0].length() - 1);
    }

    /**
     * Find host in url.
     * @param urlSplit url
     */
    private void findHost(String[] urlSplit) {
        if (urlSplit[2].split("\\.").length < 2) {
            isCorrectUrl = false;
        }

        if (urlSplit[2].split(":")[0] != null) {
            host = urlSplit[2].split(":")[0];
            host = removeFragmentFromElement(removeQueryFromElement(host));
            return;
        }
        host = removeFragmentFromElement(removeQueryFromElement(urlSplit[2]));
    }

    /**
     * Find port for url.
     * @param urlSplit url
     */
    private void findPort(String[] urlSplit) {
        if (urlSplit[2].split(":").length == 2) {
            port = Integer.parseInt(urlSplit[2].split(":")[1]);
            return;
        }
        port = portResolver.getPort(protocol);
    }

    /**
     * Remove query from url.
     * @param url url
     * @return url without query
     */
    private String removeQueryFromUrl(String url){
        if (url.split("\\?").length > 0){
            return url.split("\\?")[0];
        }
        return url;
    }

    /**
     * Check if queries are same.
     * @param query query
     * @return true/false
     */
    private Boolean queryCompare(String query){
        return querySort(query).equals(this.query);
    }

    /**
     * Returns raw url.
     * @return url
     */
    @Override
    public String getAsRawString() {
        return rawUrl;
    }

    /**
     * Returns canonical url.
     * @return url
     */
    @Override
    public String getAsString() {
        return canonicalUrl;
    }

    /**
     * Returns host.
     * @return url
     */
    @Override
    public String getHost() {
        return host;
    }

    /**
     * Returns protocol.
     * @return protocol
     */
    @Override
    public String getProtocol() {
        return protocol;
    }

    /**
     * Returns port.
     * @return port
     */
    @Override
    public int getPort() {
        return port;
    }

    /**
     * Returns path.
     * @return path
     */
    @Override
    public String getPath() {
        return path;
    }

    /**
     * Returns query.
     * @return query
     */
    @Override
    public String getQuery() {
        return query;
    }

    /**
     * Returns fragment.
     * @return fragment
     */
    @Override
    public String getFragment() {
        return fragment;
    }

    /**
     * Return True if urls are same, false otherwise.
     * @param url another URL
     * @return True/False
     */
    @Override
    public boolean isSameAs(SmartUrl url) {
        return removeQueryFromUrl(canonicalUrl).equals(removeQueryFromUrl(url.getAsString()))
                && queryCompare(url.getQuery()) && fragment.equals(url.getFragment());
    }

    /**
     * Return True if urls are same, false otherwise.
     * @param url another URL
     * @return True/False
     */
    @Override
    public boolean isSameAs(String url) {
        return url.equals(rawUrl);
    }
}
